import numpy as np
from math import pi
from ngsolve import *
from ngsolve.internal import visoptions
import ngsolve.internal
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry

def Generate2DMesh(maxh):

    geometry = SplineGeometry()
    
    # point coordinates ...
    pnts = [ (-0.5,-0.5), (0,-0.5), (0,0.5), (-0.5,0.5), (0.5,-0.5), (0.5,0.5)]
    pnums = [geometry.AppendPoint(*p) for p in pnts]
    
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    geometry.Append(["line", pnums[0], pnums[1]], bc="fluid", leftdomain=1, rightdomain=0)
    geometry.Append(["line", pnums[1], pnums[2]], bc="interface", leftdomain=1, rightdomain=2)
    geometry.Append(["line", pnums[2], pnums[3]], bc="fluid", leftdomain=1, rightdomain=0)
    geometry.Append(["line", pnums[3], pnums[0]], bc="fluid", leftdomain=1, rightdomain=0)
    geometry.Append(["line", pnums[1], pnums[4]], bc="solid", leftdomain=2, rightdomain=0)
    geometry.Append(["line", pnums[4], pnums[5]], bc="solid", leftdomain=2, rightdomain=0)
    geometry.Append(["line", pnums[5], pnums[2]], bc="solid", leftdomain=2, rightdomain=0)

    geometry.SetMaterial(1,"fluid")
    geometry.SetMaterial(2,"solid")

    mesh = Mesh(geometry.GenerateMesh(maxh=maxh))
    print(mesh.GetMaterials())
    return mesh

order=4
mesh_size=0.1
# Generate mesh
mesh = Generate2DMesh(mesh_size)

def spaces(order):
    U_s = VectorL2(mesh, order=order, complex=True, definedon="solid")
    F_s = FacetFESpace(mesh, order=order, complex=True, definedon="solid")

    U_f = VectorL2(mesh, order=order, complex=True, definedon="fluid")
    F_f = FacetFESpace(mesh, order=order, complex=True, definedon="fluid")

    Q = L2(mesh, order=order, complex=True)

    fes = FESpace([Q, U_f, U_s, F_f, F_f, F_s, F_s])

    (p_f, u_f, u_s, uhatx_f, uhaty_f, uhatx_s, uhaty_s), \
    (q_f, v_f, v_s, vhatx_f, vhaty_f, vhatx_s, vhaty_s) = fes.TnT()

    print ("ndof-V:", U_s.ndof, ", ndof-Mu:", F_f.ndof, ", ndof-Q:", Q.ndof)

    uhat_s = CoefficientFunction( (uhatx_s, uhaty_s) )
    vhat_s = CoefficientFunction( (vhatx_s, vhaty_s) )

    uhat_tr_s = CoefficientFunction( (uhatx_s.Trace(), uhaty_s.Trace()) )
    vhat_tr_s = CoefficientFunction( (vhatx_s.Trace(), vhaty_s.Trace()) )

    uhat_f = CoefficientFunction( (uhatx_f, uhaty_f) )
    vhat_f = CoefficientFunction( (vhatx_f, vhaty_f) )

    uhat_tr_f = CoefficientFunction( (uhatx_f.Trace(), uhaty_f.Trace()) )
    vhat_tr_f = CoefficientFunction( (vhatx_f.Trace(), vhaty_f.Trace()) )

    return fes, p_f, q_f, u_f, v_f, uhat_f, vhat_f, uhat_tr_f, vhat_tr_f, \
                          u_s, v_s, uhat_s, vhat_s, uhat_tr_s, vhat_tr_s

fes, p_f, q_f, u_f, v_f, uhat_f, vhat_f, uhat_tr_f, vhat_tr_f, \
u_s, v_s, uhat_s, vhat_s, uhat_tr_s, vhat_tr_s = spaces(order)

mat = mesh.GetMaterials()
bound = mesh.GetBoundaries()
geo = mesh.ngmesh.GetGeometry()
opt = mesh.ngmesh.OptimizeVolumeMesh()
Draw( mesh)
print(mat, bound)


# Solid Data
# Wavenumber & source
omega_s = 20
rho = 1
E, nu = 1, 0.3
mu  = E / 2 / (1+nu)
lamb = E * nu / ((1+nu)*(1-2*nu))

print(lamb, mu)

kp = omega_s * (sqrt(rho/(2*mu + lamb)))
ks = omega_s * (sqrt(rho/mu))

# stabilization parameter - solid
h = specialcf.mesh_size
alpha_s = 2
beta_s = alpha_s*(order+1)**2/h

# Fluid Data
# Wavenumber & source
k_f = omega_s
theta0 = 30
theta0 = theta0 * (pi/180)

# stabilization parameter - fluid
h = specialcf.mesh_size
n = specialcf.normal(mesh.dim)

alpha_f = 2
beta_f = alpha_f*(order+1)**2/h


print(k_f, kp, ks)

theta_kf = theta0
theta_kp = np.arcsin(k_f * sin(theta_kf) / kp)

theta_ks = np.arcsin(k_f * sin(theta_kf) / ks)

alpha_f =  0.0005702726491706206
alpha_p =  -0.07152201143963699j
alpha_s =  0.04473165583206899j

print("alpha_f = ", alpha_f)
print("alpha_p = ", alpha_p)
print("alpha_s = ", alpha_s)


# Exact Solution
p_exact = exp((1j*k_f)*(x*(cos(theta_kf)) + y*(sin(theta_kf)))) + alpha_f * exp((1j*k_f)*(-x*(cos(theta_kf)) + y*(sin(theta_kf))))

p_exact_pt1 = exp((1j*k_f)*(x*(cos(theta_kf)) + y*(sin(theta_kf))))
p_exact_pt2 = alpha_f * exp((1j*k_f)*(-x*(cos(theta_kf)) + y*(sin(theta_kf))))

coef1 = CoefficientFunction( (cos(theta_kp),sin(theta_kp)) )
coef2 = CoefficientFunction( (-sin(theta_ks),cos(theta_ks)) )

u_exact = alpha_p * coef1 * exp((1j*kp)*(x*(cos(theta_kp)) + y*(sin(theta_kp)))) +\
          alpha_s *  coef2 * exp((1j*ks)*(x*(cos(theta_ks)) + y*(sin(theta_ks))))


# Fluid Domain
condense = True

jump_u_f = u_f-uhat_f
jump_v_f = v_f-vhat_f


vec_r1 = CoefficientFunction( (cos(theta_kf),sin(theta_kf)) )
r1 = InnerProduct(vec_r1,n)

vec_r2 = CoefficientFunction( (-cos(theta_kf),sin(theta_kf)) )
r2 = alpha_f * InnerProduct(vec_r2,n)

delta1 = 0.5/k_f**2
delta2 =-0.5
delta3 = 0.5/k_f**2

curl_u_f = grad(u_f)[1,0] - grad(u_f)[0,1]
curl_v_f = grad(v_f)[1,0] - grad(v_f)[0,1]

divu_f = grad(u_f)[1,0] + grad(u_f)[0,1]
divv_f = grad(v_f)[1,0] + grad(v_f)[0,1]


a = BilinearForm(fes, condense=condense)

dS = dx(element_boundary=True, definedon=mesh.Materials("fluid"))
a += InnerProduct(u_f, v_f)*dx("fluid") - InnerProduct(p_f, div(v_f))*dx("fluid") + k_f**2*InnerProduct(p_f, q_f)*dx("fluid") + \
    - InnerProduct(q_f, div(u_f))*dx("fluid") + beta_f*InnerProduct(jump_u_f, jump_v_f) * dS + \
    InnerProduct(p_f, jump_v_f*n)*dS + InnerProduct(q_f, jump_u_f*n)*dS

a += (1j/k_f)*InnerProduct(uhat_tr_f*n, vhat_tr_f*n) * ds(definedon=mesh.Boundaries("fluid"))

a += delta1*InnerProduct((div(u_f) - k_f**2*p_f), (div(v_f) - k_f**2*q_f))*dx("fluid") +\
    delta2*InnerProduct((u_f + grad(p_f)), (v_f + grad(q_f)))*dx("fluid")  +\
    delta3*InnerProduct(curl_u_f, curl_v_f)*dx("fluid")


f = LinearForm(fes)
f += SymbolicLFI(0*q_f)
f += (1j/k_f)*k_f*1j*(r1 + 1)* (vhat_tr_f*n) * p_exact_pt1 * ds(definedon=mesh.Boundaries("fluid"))
f += (1j/k_f)*k_f*1j*(r2 + 1)* (vhat_tr_f*n) * p_exact_pt2 * ds(definedon=mesh.Boundaries("fluid"))


# Solid Domain
jump_u_s = u_s-uhat_s
jump_v_s = v_s-vhat_s

def sigman(x,y):
    "Inhomogeneous Robin Boundary Condition"

    mip = mesh(0.1, 0.5)

    ekp = exp(complex(0, 1) * kp *( x * cos(theta_kp) + y * sin(theta_kp)))
    eks = exp(complex(0, 1) * ks *( x * cos(theta_ks) + y * sin(theta_ks)))
    
    u11 = alpha_p * complex(0, 1) * kp * cos(theta_kp)**2 * ekp - alpha_s * complex(0, 1) * ks * sin(theta_ks) * cos(theta_ks) * eks
    u22 = alpha_p * complex(0, 1) * kp * sin(theta_kp)**2 * ekp + alpha_s * complex(0, 1) * ks * sin(theta_ks) * cos(theta_ks) * eks
    aux1 = alpha_p * complex(0, 1) * kp * sin(theta_kp) * cos(theta_kp)  * ekp - alpha_s * complex(0, 1) * ks * sin(theta_ks)**2 * eks
    aux2 = alpha_p * complex(0, 1) * kp * sin(theta_kp) * cos(theta_kp)  * ekp + alpha_s * complex(0, 1) * ks * cos(theta_ks)**2 * eks
    u12 = (aux1 + aux2)/2
    u21 = u12

    gradSu = CoefficientFunction((u11,u12,u21,u22),dims=(2,2))
    divu = u11 + u22
    sigman = 2*mu * (gradSu ) + lamb * divu * Id(mesh.dim)
    
    sigman = -sigman
    
    return sigman

# Solid Domain
matA = CoefficientFunction((kp,0,0,ks),dims=(2,2))

def epsilon(u):
    return 0.5*(grad(u) + grad(u).trans)

def sigma(u):
    return lamb*Trace(epsilon(u))*Id(mesh.dim) + 2*mu*epsilon(u)


def rbc_react():
    ekp = exp(complex(0, 1) * kp *( x * cos(theta_kp) + y * sin(theta_kp)))
    eks = exp(complex(0, 1) * ks *( x * cos(theta_ks) + y * sin(theta_ks)))
    a11 = alpha_p * complex(0, 1) * cos(theta_kp) * ekp - alpha_s * complex(0, 1) * sin(theta_ks) * eks
    a22 = alpha_p * complex(0, 1) * sin(theta_kp) * ekp + alpha_s * complex(0, 1) * cos(theta_ks) * eks
    react = CoefficientFunction((a11,a22))

    return react


force = CoefficientFunction( (0,0) )


dS = dx(element_boundary=True, definedon=mesh.Materials("solid"))
a += InnerProduct(sigma(u_s), epsilon(v_s)) * dx("solid") - rho * omega_s**2 * InnerProduct(u_s, v_s) * dx("solid") +\
     beta_s*InnerProduct(jump_u_s,jump_v_s) * dS + \
    -InnerProduct(sigma(u_s)*n, jump_v_s) * dS - InnerProduct(sigma(v_s)*n, jump_u_s) * dS
a += -1j*matA*uhat_tr_s*vhat_tr_s* ds(definedon=mesh.Boundaries("solid"))

f += SymbolicLFI(force*v_s)
f += -InnerProduct(matA*rbc_react(),vhat_tr_s) * ds(definedon=mesh.Boundaries("solid"))
f += -InnerProduct((sigman(x,y))*n,vhat_tr_s)* ds(definedon=mesh.Boundaries("solid"))


# Interface fluid-solid (lamb_s)
interfaceTerm = p_f*(v_f - vhat_tr_f)*n + \
                q_f*(u_f - rho*omega_s**2*uhat_tr_s)*n + \
     beta_f*(u_f - rho*omega_s**2*uhat_tr_s)*n*(v_f - vhat_tr_f)*n

interface_int = interfaceTerm *ds(definedon=mesh.Boundaries("interface"))
a += interface_int
   

a.Assemble()
f.Assemble()

gfu = GridFunction(fes)

# solve linear system
if not condense:
    inv = a.mat.Inverse(fes.FreeDofs(), "umfpack")
    gfu.vec.data = inv * f.vec
else:
    f.vec.data += a.harmonic_extension_trans * f.vec 
    
    inv = a.mat.Inverse(fes.FreeDofs(True), "sparsecholesky")
    gfu.vec.data = inv * f.vec
    
    gfu.vec.data += a.harmonic_extension * gfu.vec
    gfu.vec.data += a.inner_solve * f.vec

Draw (gfu.components[0], mesh, "p")
Draw (gfu.components[1][0], mesh, "u_f")
Draw (gfu.components[2][0], mesh, "u_s")


p_exact = CoefficientFunction(p_exact)

p_h = CoefficientFunction(gfu.components[0])

l2norm = sqrt (Integrate ( (p_h-p_exact)*Conj(p_h-p_exact), mesh))
#Error analysis
print ("Pres L2-error:", l2norm)


folder = "results/"
filename = "fsiSol_order_" + str(order)
fileName = folder + filename

vtk = VTKOutput(ma=mesh, coefs=[gfu.components[0].real], names = ["p_real"], filename= fileName + "_p_real", subdivision=3)
vtk.Do()
vtk = VTKOutput(ma=mesh, coefs=[gfu.components[0].imag],  names=["p_imag"], filename= fileName + "_p_imag", subdivision=3)
vtk.Do()

vtk = VTKOutput(ma=mesh, coefs=[gfu.components[1][0].real], names = ["uf_real"], filename= fileName + "_uf_real", subdivision=3)
vtk.Do()
vtk = VTKOutput(ma=mesh, coefs=[gfu.components[1][0].imag],  names=["uf_imag"], filename= fileName + "_uf_imag", subdivision=3)
vtk.Do()

vtk = VTKOutput(ma=mesh, coefs=[gfu.components[2][0].real], names = ["us_real"], filename= fileName + "_us_real", subdivision=3)
vtk.Do()
vtk = VTKOutput(ma=mesh, coefs=[gfu.components[2][0].imag],  names=["us_imag"], filename= fileName + "_us_imag", subdivision=3)
vtk.Do()

print ("End Program")
